  // @file itf2_image.cc
  // @brief Image data structures and functions.
  // @author Mamoru Kaminaga
  // @date 2017-12-14 10:34:25
  // Copyright 2017 Mamoru Kaminaga
  // Released under the MIT license
  // http://opensource.org/licenses/mit-license.php
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <wchar.h>
#include <png.h>
#include <algorithm>
#include <cmath>
#include <memory>
#include <vector>
#include "./itf2_image.h"
#include "./util.h"

#define RGB_TO_PALETTE(r, g, b) (((r) << 16) | ((g) << 8) | (b))
#define PALETTE_TO_RED(c)       ((c) >> 16)
#define PALETTE_TO_GREEN(c)     (((c) >> 8) & 0xff)
#define PALETTE_TO_BLUE(c)      ((c) & 0xff)
#define INT_TRUE                (1)
#define INT_FALSE               (0)
#define IM0_HEADER_SIZE         (9)
#define PNG_HEADER_SIZE         (8)

struct ITF2Image::Pimpl {
  png_byte color_type;
  int width;
  int height;
  int pixel_num;
  int color_num;
  int bit_num;
  std::vector<int> palette;
  std::vector<int> indices;
  std::vector<int> color_is_used;

  void Clear() {
    color_type = 0;
    width = 0;
    height = 0;
    pixel_num = 0;
    color_num = 0;
    bit_num = 0;
    palette.clear();
    indices.clear();
    color_is_used.clear();
  }

  bool RemoveNonUsedColors() {
    // Array to store new image data.
    std::vector<int> new_palette;
    std::vector<int> new_indices(pixel_num);

    // For each palette colors, check used or not.
    int new_color_id = 0;
    for (int i = 0; i < color_num; ++i) {
      if (color_is_used[i] == INT_TRUE) {
        // If the color is used, add the color to the new palette.
        new_palette.push_back(palette[i]);
        new_color_id = new_palette.size() - 1;
        color_is_used[new_color_id] = INT_TRUE;
        // The new color id is reflected to pixel color index array.
        for (int j = 0; j < pixel_num; ++j) {
          if (indices[j] == i) {
            new_indices[j] = new_color_id;
          }
        }
      }
    }

    // The result is feed bucked.
    palette.clear();
    indices.clear();
    color_num = new_palette.size();
    palette.resize(color_num);
    indices.resize(pixel_num);
    std::copy(new_palette.begin(), new_palette.end(), palette.begin());
    std::copy(new_indices.begin(), new_indices.end(), indices.begin());

    // The bits are checked.
    if (color_num <= 2) {
      bit_num = 1;
    } else if (color_num <= 4) {
      bit_num = 2;
    } else if (color_num <= 16) {
      bit_num = 4;
    } else if (color_num <= 256) {
      bit_num = 8;
    } else {
      fwprintf(stderr, L"Error... Too many colors\n");
      return false;
    }
    return true;
  }
};
ITF2Image::ITF2Image() : pimpl_(new ITF2Image::Pimpl()) { }
ITF2Image::~ITF2Image() = default;

bool ITF2Image::ReadPNGFile(const wchar_t* file_name) {
  pimpl_->Clear();

  // The file is opened and the signature is checked.
  FILE* fp = Fopen(file_name, L"rb");
  if (fp == nullptr) {
    fwprintf(stderr, L"Error... File %ls cannot be opened\n", file_name);
    return false;
  }

  // The file signature is checked.
  uint8_t header[PNG_HEADER_SIZE] = {0};
  fread(header, 1, PNG_HEADER_SIZE, fp);
  int is_png = !png_sig_cmp(header, 0, PNG_HEADER_SIZE);
  if (!is_png) {
    fwprintf(stderr, L"Error... File %ls not PNG\n", file_name);
    return false;
  }

  // png_struct initialization.
  png_structp png_ptr = png_create_read_struct(
      PNG_LIBPNG_VER_STRING, nullptr, nullptr, nullptr);
  if (png_ptr == nullptr) {
    fwprintf(stderr, L"Error... png_create_read_struct failed\n");
    return false;
  }

  // png_info initialization for data chunk before image data.
  png_infop info_ptr = png_create_info_struct(png_ptr);
  if (info_ptr == nullptr) {
    fwprintf(stderr, L"Error... png_destroy_read_struct failed\n");
    goto ERROR;
  }

  // setjmp, longjmp
  if (setjmp(png_jmpbuf(png_ptr))) {
    goto ERROR;
  }

  // The file pointer is set.
  png_init_io(png_ptr, fp);
  png_set_sig_bytes(png_ptr, PNG_HEADER_SIZE);

  // The file is read with high level API.
  png_read_png(
      png_ptr, info_ptr,
      PNG_TRANSFORM_PACKING | PNG_TRANSFORM_STRIP_16,
      nullptr);
  pimpl_->width = png_get_image_width(png_ptr, info_ptr);
  pimpl_->height = png_get_image_height(png_ptr, info_ptr);
  pimpl_->bit_num = png_get_bit_depth(png_ptr, info_ptr);

  // Data buffer is resized.
  pimpl_->pixel_num = pimpl_->width * pimpl_->height;
  pimpl_->indices.resize(pimpl_->pixel_num);

  // Get pixels.
  png_bytepp rows = png_get_rows(png_ptr, info_ptr);
  pimpl_->color_type = png_get_color_type(png_ptr, info_ptr);
  switch (pimpl_->color_type) {
    case PNG_COLOR_TYPE_PALETTE:
      {
        // Color palette is acquired.
        png_colorp pal = nullptr;
        png_get_PLTE(png_ptr, info_ptr, &pal, &pimpl_->color_num);
        pimpl_->palette.resize(pimpl_->color_num);
        pimpl_->color_is_used.resize(pimpl_->color_num);
        for (int i = 0; i < pimpl_->color_num; ++i) {
          pimpl_->palette[i] = RGB_TO_PALETTE(
              pal[i].red, pal[i].green, pal[i].blue);
        }
        // Color indices are acquired.
        int pixel_id = 0;
        int color_id = 0;
        png_bytep row = nullptr;
        for (int y = 0; y < pimpl_->height; ++y) {
          row = rows[y];
          for (int x = 0; x < pimpl_->width; ++x) {
            pixel_id = y * pimpl_->width + x;
            color_id = *row;
            row += sizeof(png_byte);
            pimpl_->indices[pixel_id] = color_id;

            // The color use state flag is set.
            pimpl_->color_is_used[color_id] = INT_TRUE;
          }
        }
      }
      break;
    case PNG_COLOR_TYPE_RGB:
      {
        // Full color data to indexed color.
        png_bytep row = nullptr;
        int red = 0;
        int green = 0;
        int blue = 0;
        int c = 0;
        int color_id = 0;
        int pixel_id = 0;
        pimpl_->palette.clear();
        for (int y = 0; y < pimpl_->height; ++y) {
          for (int x = 0; x < pimpl_->width; ++x) {
            row = rows[y];
            pixel_id = y * pimpl_->width + x;
            // Red
            red = *row;
            row += sizeof(png_byte);
            // Green
            green = *row;
            row += sizeof(png_byte);
            // Blue
            blue = *row;
            row += sizeof(png_byte);
            // TODO(Mamoru): O(N) for each loop is not good way for search.
            // Replace this with more efficient algorithm.
            // Find color pixel_id for the color.
            c = RGB_TO_PALETTE(red, green, blue);
            auto it = std::find(
                pimpl_->palette.begin(), pimpl_->palette.end(), c);
            if (it == pimpl_->palette.end()) {
              pimpl_->palette.push_back(c);
            }
            color_id = std::distance(pimpl_->palette.begin(), it);
            pimpl_->indices[pixel_id] = color_id;

            // The color use state flag is set.
            pimpl_->color_is_used[color_id] = INT_TRUE;
          }
        }
      }
      break;
    default:
      fwprintf(stderr, L"Error... This PNG format is not supported\n");
      break;
  }

  fclose(fp);
  return true;

ERROR:
  fclose(fp);
  png_destroy_read_struct(&png_ptr, &info_ptr, nullptr);
  return false;
}
bool ITF2Image::ReadTextFile(const wchar_t* file_name, TEXTFORMAT format) {
  pimpl_->Clear();

  FILE* fp = Fopen(file_name, L"r");
  if (fp == nullptr) {
    fwprintf(stderr, L"Error... File %ls is cannot be opened\n", file_name);
    return false;
  }
  switch (format) {
    case TEXTFORMAT_DEC:
    case TEXTFORMAT_HEX:
      {
        RWMODE mode;
        if (format == TEXTFORMAT_DEC) {
          mode = RWMODE_DEC;
        } else {
          mode = RWMODE_HEX;
        }
        // The file is opened and the signature is read.
        int sign[3] = {0};
        sign[0] = ReadByte(fp, mode);
        sign[1] = ReadByte(fp, mode);
        sign[2] = ReadByte(fp, mode);
        if ((sign[0] != L'I') || (sign[1] != L'M') || (sign[2] != L'0')) {
          fwprintf(stderr, L"Error... Signature is not IM0\n");
          goto ERROR;
        }
        // The image total size is acquired.
        int total_size = ReadByte(fp, mode);
        if (total_size == 0) {
          fwprintf(stderr, L"Error... Total size is 0\n");
          goto ERROR;
        }
        // The image address size is acquired.
        int offset_to_image = ReadByte(fp, mode);
        if (offset_to_image == 0) {
          fwprintf(stderr, L"Error... Image address is 0\n");
          goto ERROR;
        }
        // The image width is acquired.
        pimpl_->width = ReadByte(fp, mode);
        if (pimpl_->width == 0) {
          fwprintf(stderr, L"Error... Image width is 0\n");
          goto ERROR;
        }
        // The image height is acquired.
        pimpl_->height = ReadByte(fp, mode);
        if (pimpl_->height == 0) {
          fwprintf(stderr, L"Error... Image height is 0\n");
          goto ERROR;
        }
        // The image bit num is acquired.
        pimpl_->bit_num = ReadByte(fp, mode);
        if ((pimpl_->bit_num != 1) && (pimpl_->bit_num != 2) &&
            (pimpl_->bit_num != 4) && (pimpl_->bit_num != 8)) {
          fwprintf(stderr, L"Error... Bit num is not 1, 2, 4 or 8\n");
          goto ERROR;
        }
        // The color num num is acquired.
        pimpl_->color_num = ReadByte(fp, mode);
        if (pimpl_->color_num == 0) {
          fwprintf(stderr, L"Error... Color num is 0\n");
          goto ERROR;
        }
        // The color array is acquired.
        pimpl_->palette.resize(pimpl_->color_num);
        pimpl_->color_is_used.resize(pimpl_->color_num);
        int red = 0;
        int green = 0;
        int blue = 0;
        for (int i = 0; i < pimpl_->color_num; ++i) {
          red = ReadByte(fp, mode);
          green = ReadByte(fp, mode);
          blue = ReadByte(fp, mode);
          pimpl_->palette[i] = RGB_TO_PALETTE(red, green, blue);
        }
        // The pixel array is acquired.
        pimpl_->pixel_num = pimpl_->width * pimpl_->height;
        pimpl_->indices.resize(pimpl_->pixel_num);
        int data = 0;
        for (int i = 0; i < pimpl_->pixel_num; /* No op */ ) {
          data = ReadByte(fp, mode);
          switch (pimpl_->bit_num) {
            case 1:
              for (int j = 0; j < 8; ++j) {
                if (i < pimpl_->pixel_num) {
                  pimpl_->indices[i] = (data >> (7 - j)) & 0x01;
                }
                ++i;
              }
              break;
            case 2:
              for (int j = 0; j < 4; ++j) {
                if (i < pimpl_->pixel_num) {
                  pimpl_->indices[i] = (data >> (6 - j * 2)) & 0x03;
                }
                ++i;
              }
              break;
            case 4:
              for (int j = 0; j < 2; ++j) {
                if (i < pimpl_->pixel_num) {
                  pimpl_->indices[i] = (data >> (4 - j * 4)) & 0x0f;
                }
                ++i;
              }
              break;
            case 8:
              pimpl_->indices[i] = data;
              ++i;
              break;
            default:
              // This case never comes.
              break;
          }
        }
      }
      break;
    default:
      // No implementation.
      return false;
  }
  fclose(fp);
  return true;

ERROR:
  fclose(fp);
  return false;
}
bool ITF2Image::WritePNGFile(const wchar_t* file_name) {
  if ((pimpl_->color_num == 0) || (pimpl_->pixel_num == 0)) return false;

  // Local variable fp must be initialized before jump to the ERROR label.
  FILE* fp = nullptr;
  png_bytepp rows = nullptr;

  // The PNG write struct is created.
  png_structp png_ptr = png_create_write_struct(
      PNG_LIBPNG_VER_STRING, nullptr, nullptr, nullptr);
  if (png_ptr == nullptr) {
    fwprintf(stderr, L"Error... png_create_write_struct failed\n");
    goto ERROR;
  }

  // The PNG info struct is created.
  png_infop info_ptr = nullptr;
  info_ptr = png_create_info_struct(png_ptr);
  if (info_ptr == nullptr) {
    fwprintf(stderr, L"Error... png_create_info_struct failed\n");
    goto ERROR;
  }

  // setjmp/longjmp
  if (setjmp(png_jmpbuf(png_ptr))) {
    goto ERROR;
  }

  // The file is opened.
  fp = Fopen(file_name, L"wb");
  if (fp == nullptr) {
    fwprintf(stderr, L"Error... File %ls is cannot be opened\n", file_name);
    return false;
  }
  png_init_io(png_ptr, fp);

  // Initialize info structure.
  // PNG image is drawn with 8bit index color.
  png_set_IHDR(png_ptr, info_ptr, pimpl_->width, pimpl_->height,
      8, PNG_COLOR_TYPE_PALETTE, PNG_INTERLACE_NONE,
      PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
  rows = (png_bytepp) png_malloc(
      png_ptr, sizeof(png_bytep) * pimpl_->height);
  if (rows == nullptr) {
    fwprintf(stderr, L"Error... png_malloc failed\n");
    goto ERROR;
  }
  png_set_rows(png_ptr, info_ptr, rows);
  memset(rows, 0, sizeof(png_bytep) * pimpl_->height);
  for (int y = 0; y < pimpl_->height; ++y) {
    rows[y] = (png_bytep) png_malloc(png_ptr, sizeof(png_byte) * pimpl_->width);
    if (rows[y] == nullptr) {
      fwprintf(stderr, L"Error... png_malloc failed\n");
      goto ERROR;
    }
  }

  // The color palette is set.
  png_colorp pal = (png_colorp) png_malloc(
      png_ptr, sizeof(png_color) * pimpl_->color_num);
  if (pal == nullptr) {
    fwprintf(stderr, L"Error... png_malloc failed\n");
    goto ERROR;
  }
  for (int i = 0; i < pimpl_->color_num; ++i) {
    pal[i].red = static_cast<png_byte>(PALETTE_TO_RED(pimpl_->palette[i]));
    pal[i].green = static_cast<png_byte>(PALETTE_TO_GREEN(pimpl_->palette[i]));
    pal[i].blue = static_cast<png_byte>(PALETTE_TO_BLUE(pimpl_->palette[i]));
  }
  png_set_PLTE(png_ptr, info_ptr, pal, pimpl_->color_num);
  png_free(png_ptr, pal);

  // Pixels are set.
  int index = 0;
  png_bytep row = nullptr;
  for (int y = 0; y < pimpl_->height; ++y) {
    row = rows[y];
    for (int x = 0; x < pimpl_->width; ++x) {
      index = y * pimpl_->width + x;
      *row = static_cast<png_byte>(pimpl_->indices[index]);
      row += sizeof(png_byte);
    }
  }

  // The PNG file is written.
  png_write_png(png_ptr, info_ptr, PNG_TRANSFORM_IDENTITY, nullptr);

  // Finalization.
  for (int y = 0; y < pimpl_->height; ++y) {
    png_free(png_ptr, rows[y]);
  }
  png_free(png_ptr, rows);
  png_destroy_write_struct(&png_ptr, &info_ptr);
  fclose(fp);
  return true;

ERROR:
  if (rows != nullptr) {
    for (int y = 0; y < pimpl_->height; ++y) {
      if (rows[y] != nullptr) png_free(png_ptr, rows[y]);
    }
    png_free(png_ptr, rows);
  }
  png_destroy_write_struct(&png_ptr, &info_ptr);
  if (fp != nullptr) fclose(fp);
  return false;
}
bool ITF2Image::WriteTextFile(const wchar_t* file_name, TEXTFORMAT format) {
  if ((pimpl_->color_num == 0) || (pimpl_->pixel_num == 0)) return false;

  // Data size to be smaller than original PNG image.
  pimpl_->RemoveNonUsedColors();

  const int palette_size = pimpl_->color_num * 3;
  const int offset_to_image = IM0_HEADER_SIZE + palette_size;
  const int total_size = GetImageDataSize();

  // The binary buffer is created.
  std::vector<uint8_t> buffer(total_size);
  buffer[0] = 'I';
  buffer[1] = 'M';
  buffer[2] = '0';
  buffer[3] = static_cast<uint8_t>(total_size);
  buffer[4] = static_cast<uint8_t>(offset_to_image);
  buffer[5] = static_cast<uint8_t>(pimpl_->width);
  buffer[6] = static_cast<uint8_t>(pimpl_->height);
  buffer[7] = static_cast<uint8_t>(pimpl_->bit_num);
  buffer[8] = static_cast<uint8_t>(pimpl_->color_num);
  for (int i = 0; i < pimpl_->color_num; ++i) {
    buffer[IM0_HEADER_SIZE + i * 3 + 0] =
      static_cast<uint8_t>(PALETTE_TO_RED(pimpl_->palette[i]));
    buffer[IM0_HEADER_SIZE + i * 3 + 1] =
      static_cast<uint8_t>(PALETTE_TO_GREEN(pimpl_->palette[i]));
    buffer[IM0_HEADER_SIZE + i * 3 + 2] =
      static_cast<uint8_t>(PALETTE_TO_BLUE(pimpl_->palette[i]));
  }
  int byte_cnt = 0;
  uint8_t data = 0;
  for (int i = 0; i < pimpl_->pixel_num; /* No op */ ) {
    data = 0;
    switch (pimpl_->bit_num) {
      case 1:
        for (int j = 0; j < 8; ++j) {
          if (i < pimpl_->pixel_num) {
            data |= static_cast<uint8_t>(pimpl_->indices[i] << (7 - j));
          } else {
            break;
          }
          ++i;
        }
        break;
      case 2:
        for (int j = 0; j < 4; ++j) {
          if (i < pimpl_->pixel_num) {
            data |= static_cast<uint8_t>(pimpl_->indices[i] << (6 - j * 2));
          } else {
            break;
          }
          ++i;
        }
        break;
      case 4:
        for (int j = 0; j < 2; ++j) {
          if (i < pimpl_->pixel_num) {
            data |= static_cast<uint8_t>(pimpl_->indices[i] << (4 - j * 4));
          } else {
            break;
          }
          ++i;
        }
        break;
      case 8:
        data = static_cast<uint8_t>(pimpl_->indices[i]);
        ++i;
        break;
      default:
        // This case never comes.
        return false;
    }
    buffer[offset_to_image + byte_cnt] = data;
    ++byte_cnt;
  }

  // The file is opened.
  FILE* fp = Fopen(file_name, L"w");
  if (fp == nullptr) {
    fwprintf(stderr, L"Error... File %ls cannot be opened\n", file_name);
    return false;
  }

  switch (format) {
    case TEXTFORMAT_DEC:
    case TEXTFORMAT_HEX:
      {
        RWMODE mode;
        if (format == TEXTFORMAT_DEC) {
          mode = RWMODE_DEC;
        } else {
          mode = RWMODE_HEX;
        }
        // The data arraignment for decimal and hexadecimal is same.
        for (int i = 0; i < static_cast<int>(total_size); ++i) {
          WriteByte(fp, static_cast<int>(buffer[i]), mode);
        }
      }
      break;
    case TEXTFORMAT_CMD:
      {
        // Time is acquired.
        struct tm local_time;
        GetLocalTime(&local_time);
        // Command file header is generated.
        fwprintf(fp,
            L"\"CMDS\",\"%02d\",\"%02d\",\"%02d\",\"%02d\",\"%02d\",\"%02d\",",
            local_time.tm_year + 1900 - 2000, local_time.tm_mon + 1,
            local_time.tm_mday, local_time.tm_hour, local_time.tm_min,
            local_time.tm_sec);
        fwprintf(fp, L"\"DLD\",\"%d\",", total_size * 2);
        // Command file binary is written.
        for (int i = 0; i < static_cast<int>(total_size); ++i) {
          WriteByte(fp, buffer[i], RWMODE_CMD);
        }
      }
      break;
    default:
      // No implementation.
      return false;
  }

  fclose(fp);
  return true;
}
bool ITF2Image::RemoveNonUsedColors() {
  pimpl_->RemoveNonUsedColors();
  return true;
}
int ITF2Image::GetImageDataSize() {
  const int palette_size = pimpl_->color_num * 3;
  const int offset_to_image = IM0_HEADER_SIZE + palette_size;
  const int image_size = static_cast<int>(
      std::ceil(
        static_cast<double>(pimpl_->pixel_num * pimpl_->bit_num / 8.0)));
  return offset_to_image + image_size;
}
bool ITF2Image::ShowImageDataSize() {
  wprintf(L"\n\n");
  wprintf(L"Total image data size: %d\n", GetImageDataSize());
  wprintf(L"\n\n");
  return true;
}
bool ITF2Image::ShowVerboseInfo() {
  // Show header info.
  wprintf(L"\n\n");
  switch (pimpl_->color_type) {
    case PNG_COLOR_TYPE_PALETTE:
      wprintf(L"color_type=PNG_COLOR_TYPE_PALETTE\n");
      break;
    case PNG_COLOR_TYPE_RGB:
      wprintf(L"color_type=PNG_COLOR_TYPE_RGB\n");
      break;
    default:
      wprintf(L"color_type=?\n");
      break;
  }
  wprintf(L"width=%d\n", pimpl_->width);
  wprintf(L"height=%d\n", pimpl_->height);
  wprintf(L"color_num=%d\n", pimpl_->color_num);
  wprintf(L"pixel_num=%d\n", pimpl_->pixel_num);
  wprintf(L"bit_num=%d\n", pimpl_->bit_num);
  wprintf(L"\n");

  // Show palette colors.
  wprintf(L"palette\n");
  wprintf(L"----\n");
  for (int i = 0; i < pimpl_->color_num; ++i) {
    wprintf(L"  %d: %02x%02x%02x\n",
        i,
        PALETTE_TO_RED(pimpl_->palette[i]),
        PALETTE_TO_GREEN(pimpl_->palette[i]),
        PALETTE_TO_BLUE(pimpl_->palette[i]));
  }
  wprintf(L"\n");

  // Show pixel color ids.
  wprintf(L"pixels\n");
  wprintf(L"----\n");
  for (int i = 0; i < pimpl_->pixel_num; ++i) {
    if (i == 0) {
      wprintf(L"  ");
    } else if ((i % pimpl_->width) == 0) {
      wprintf(L"\n  ");
    }
    wprintf(L"%-04d", pimpl_->indices[i]);
  }
  wprintf(L"\n\n");
  return true;
}
