﻿  // @file main.cc
  // @brief The main function.
  // @author Mamoru Kaminaga
  // @date 2017-09-05 19:51:09
  // Copyright 2017 Mamoru Kaminaga
  // Released under the MIT license
  // http://opensource.org/licenses/mit-license.php
#include <png.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <wchar.h>
#include <string>
#include <vector>
#include "./help.h"
#include "./itf2_image.h"
#include "./util.h"

#define ITF2IMG_VERSION    L"itf2img Version v1.0.0"
#define ITF2IMG_COPYRIGHT  L"Copyright 2017 Mamoru Kaminaga"
#define MAX_DATA_SIZE      (85)

namespace {
  enum COMMAND {
    // Info display commands.
    COMMAND_SIZE,
    COMMAND_INFO,
    // Data conversion commands.
    COMMAND_DEC2PNG,
    COMMAND_HEX2PNG,
    COMMAND_PNG2DEC,
    COMMAND_PNG2HEX,
    COMMAND_PNG2CMD,
    COMMAND_NONE,
  };
  COMMAND command = COMMAND_NONE;
  int command_argc = 0;
  bool is_verbose = false;
  std::basic_string<wchar_t> in_file;
  std::basic_string<wchar_t> out_file;
}  // namespace

#if defined(WIN32)
#define MAIN int wmain(int argc, wchar_t* wargv[])
#elif defined(UNIX)
#define MAIN int main(int argc, char* argv[])
#endif

MAIN {
  // Argument counter upper size check.
  if (argc > 6) {
    wprintf(L"Error... Too many arguments\n");
    exit(EXIT_FAILURE);
  }
#if defined(UNIX)
  // Convert command line arguments into wide characters from multi byte
  // characters.
  std::vector<wchar_t*> wargv(argc);
  std::vector<std::vector<wchar_t>> argv_buf(argc);
  for (int i = 0; i < argc; ++i) {
    argv_buf[i].resize(strlen(argv[i]) + 1);
    mbstowcs(argv_buf[i].data(), argv[i], strlen(argv[i]));
    wargv[i] = argv_buf[i].data();
  }
#endif

  // The command line arguments are processed.
  for (int i = 0; i < argc; ++i) {
    if (wcscmp(wargv[i], L"--help") == 0) {
      wprintf(HELP_MESSAGE);
      exit(EXIT_SUCCESS);
    } else if (wcscmp(wargv[i], L"--version") == 0) {
      wprintf(L"%ls\n", ITF2IMG_VERSION);
      wprintf(L"%ls\n", ITF2IMG_COPYRIGHT);
      exit(EXIT_SUCCESS);
    } else if ((wcscmp(wargv[i], L"-verbose") == 0) ||
        (wcscmp(wargv[i], L"-v") == 0)) {
      is_verbose = true;
    } else if (wcscmp(wargv[i], L"s") == 0) {
      command = COMMAND_SIZE;
      command_argc = i;
    } else if (wcscmp(wargv[i], L"i") == 0) {
      command = COMMAND_INFO;
      command_argc = i;
    } else if (wcscmp(wargv[i], L"d2p") == 0) {
      command = COMMAND_DEC2PNG;
      command_argc = i;
      goto GET_FILE_NAMES;
    } else if (wcscmp(wargv[i], L"h2p") == 0) {
      command = COMMAND_HEX2PNG;
      command_argc = i;
      goto GET_FILE_NAMES;
    } else if (wcscmp(wargv[i], L"p2d") == 0) {
      command = COMMAND_PNG2DEC;
      command_argc = i;
      goto GET_FILE_NAMES;
    } else if (wcscmp(wargv[i], L"p2h") == 0) {
      command = COMMAND_PNG2HEX;
      command_argc = i;
      goto GET_FILE_NAMES;
    } else if (wcscmp(wargv[i], L"p2c") == 0) {
      command = COMMAND_PNG2CMD;
      command_argc = i;
      goto GET_FILE_NAMES;
    }
  }

  if ((argc < 1) || (command == COMMAND_NONE)) {
    fwprintf(stderr, L"Error... Argument error\n");
    wprintf(L"Use --help option for detailed information\n");
    exit(EXIT_FAILURE);
  }

GET_FILE_NAMES:
  if (wcscmp(wargv[command_argc + 1], L"-o") == 0) {
    if (argc < command_argc + 3) {
      fwprintf(stderr, L"Error... Input file is not specified\n");
      exit(EXIT_FAILURE);
    }
    out_file = wargv[command_argc + 2];
    in_file = wargv[command_argc + 3];
  } else {
    if (argc < command_argc + 1) {
      fwprintf(stderr, L"Error... Input file is not specified\n");
      exit(EXIT_FAILURE);
    }
    if ((command == COMMAND_DEC2PNG) || (command == COMMAND_HEX2PNG)) {
      out_file = L"a.png";
    } else {
      out_file = L"a.txt";
    }
    in_file = wargv[command_argc + 1];
  }

  ITF2Image itf2img;
  switch (command) {
    case COMMAND_SIZE:
      if (!itf2img.ReadPNGFile(in_file.c_str())) {
        fwprintf(stderr, L"Failed to read %ls\n", in_file.c_str());
        exit(EXIT_FAILURE);
      }
      itf2img.RemoveNonUsedColors();
      itf2img.ShowImageDataSize();
      break;
    case COMMAND_INFO:
      if (!itf2img.ReadPNGFile(in_file.c_str())) {
        fwprintf(stderr, L"Failed to read %ls\n", in_file.c_str());
        exit(EXIT_FAILURE);
      }
      // before color reduction.
      wprintf(L"Before color reduction\n");
      wprintf(L"====\n");
      itf2img.ShowVerboseInfo();
      itf2img.RemoveNonUsedColors();
      // after color reduction.
      wprintf(L"After color reduction\n");
      wprintf(L"====\n");
      itf2img.ShowVerboseInfo();
      itf2img.ShowImageDataSize();
      break;
    case COMMAND_DEC2PNG:
      if (!itf2img.ReadTextFile(in_file.c_str(), ITF2Image::TEXTFORMAT_DEC)) {
        fwprintf(stderr, L"Failed to read %ls\n", in_file.c_str());
        exit(EXIT_FAILURE);
      }
      if (is_verbose) itf2img.ShowVerboseInfo();
      if (!itf2img.WritePNGFile(out_file.c_str())) {
        fwprintf(stderr, L"Error... Decode %ls -> %ls failed\n",
            in_file.c_str(), out_file.c_str());
        exit(EXIT_FAILURE);
      }
      if (itf2img.GetImageDataSize() > MAX_DATA_SIZE) {
        itf2img.ShowImageDataSize();
        fwprintf(stderr, L"Error... Image data is tool large");
        exit(EXIT_FAILURE);
      }
      wprintf(L"Convert %ls -> %ls\n", in_file.c_str(), out_file.c_str());
      break;
    case COMMAND_HEX2PNG:
      if (!itf2img.ReadTextFile(in_file.c_str(), ITF2Image::TEXTFORMAT_HEX)) {
        fwprintf(stderr, L"Failed to read %ls\n", in_file.c_str());
        exit(EXIT_FAILURE);
      }
      if (is_verbose) itf2img.ShowVerboseInfo();
      if (!itf2img.WritePNGFile(out_file.c_str())) {
        fwprintf(stderr, L"Error... Decode %ls -> %ls failed\n",
            in_file.c_str(), out_file.c_str());
        exit(EXIT_FAILURE);
      }
      if (itf2img.GetImageDataSize() > MAX_DATA_SIZE) {
        itf2img.ShowImageDataSize();
        fwprintf(stderr, L"Error... Image data is tool large");
        exit(EXIT_FAILURE);
      }
      wprintf(L"Convert %ls -> %ls\n", in_file.c_str(), out_file.c_str());
      break;
    case COMMAND_PNG2DEC:
      if (!itf2img.ReadPNGFile(in_file.c_str())) {
        fwprintf(stderr, L"Failed to read %ls\n", in_file.c_str());
        exit(EXIT_FAILURE);
      }
      if (is_verbose) itf2img.ShowVerboseInfo();
      if (!itf2img.WriteTextFile(out_file.c_str(), ITF2Image::TEXTFORMAT_DEC)) {
        fwprintf(stderr, L"Error... Failed to write %ls\n", out_file.c_str());
        exit(EXIT_FAILURE);
      }
      if (is_verbose) itf2img.ShowVerboseInfo();
      if (itf2img.GetImageDataSize() > MAX_DATA_SIZE) {
        itf2img.ShowImageDataSize();
        fwprintf(stderr, L"Error... Image data is tool large");
        exit(EXIT_FAILURE);
      }
      wprintf(L"Convert %ls -> %ls\n", in_file.c_str(), out_file.c_str());
      break;
    case COMMAND_PNG2HEX:
      if (!itf2img.ReadPNGFile(in_file.c_str())) {
        fwprintf(stderr, L"Failed to read %ls\n", in_file.c_str());
        exit(EXIT_FAILURE);
      }
      if (is_verbose) itf2img.ShowVerboseInfo();
      if (!itf2img.WriteTextFile(out_file.c_str(), ITF2Image::TEXTFORMAT_HEX)) {
        fwprintf(stderr, L"Error... Failed to write %ls\n", out_file.c_str());
        exit(EXIT_FAILURE);
      }
      if (is_verbose) itf2img.ShowVerboseInfo();
      if (itf2img.GetImageDataSize() > MAX_DATA_SIZE) {
        itf2img.ShowImageDataSize();
        fwprintf(stderr, L"Error... Image data is tool large");
        exit(EXIT_FAILURE);
      }
      wprintf(L"Convert %ls -> %ls\n", in_file.c_str(), out_file.c_str());
      break;
    case COMMAND_PNG2CMD:
      if (!itf2img.ReadPNGFile(in_file.c_str())) {
        fwprintf(stderr, L"Failed to read %ls\n", in_file.c_str());
        exit(EXIT_FAILURE);
      }
      if (is_verbose) itf2img.ShowVerboseInfo();
      if (!itf2img.WriteTextFile(out_file.c_str(), ITF2Image::TEXTFORMAT_CMD)) {
        fwprintf(stderr, L"Error... Failed to write %ls\n", out_file.c_str());
        exit(EXIT_FAILURE);
      }
      if (is_verbose) itf2img.ShowVerboseInfo();
      if (itf2img.GetImageDataSize() > MAX_DATA_SIZE) {
        itf2img.ShowImageDataSize();
        fwprintf(stderr, L"Error... Image data is tool large");
        exit(EXIT_FAILURE);
      }
      wprintf(L"Convert %ls -> %ls\n", in_file.c_str(), out_file.c_str());
      break;
    default:
      // This case never comes.
      break;
  }

  exit(EXIT_SUCCESS);
}
